import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.widgets import Button


def disc_mask(shape, radius):
    """\
    Helper function to generate a disc mask for Fourier filtering
    """
    q2 = compute_q(shape)[2]
    return q2 <= radius**2


def box_mask(shape, size):
    """\
    Helper function to generate a disc mask for Fourier filtering
    """
    try:
        x, y = size
    except:
        x = size
        y = x
    q0, q1, q2 = compute_q(shape)
    return (abs(q0) < y//2)*(abs(q1) < x//2) 


def gauss_mask(shape, width):
    """\
    Helper function to generate a disc mask for Fourier filtering
    """
    q2 = compute_q(shape)[2]
    return np.exp(-q2 / (width/2)**2)


def compute_q(shape):
    n0, n1 = shape
    x0, x1 = np.indices(shape)
    q0 = (x0 + (n0+1)//2) % n0 - (n0+1)//2
    q1 = (x1 + (n1+1)//2) % n1 - (n1+1)//2
    q2 = q0**2 + q1**2
    return (q0, q1, q2)


class FExplorer(object):

    def __init__(self, img, mask=None, cmap=None, useblit=True):
        # Store initial image
        self.img = img
        self.shape = img.shape
        self.fimg = np.fft.fftn(img)

        if mask is None:
            self.mask = np.ones(img.shape, dtype=float)
        else:
            self.mask = mask.astype(float)

        self.symmetrize_mask()

        # Prepare plotting
        self.fig = plt.figure(1)
        self.fig.set_facecolor('white')
        self.fig.canvas.set_window_title('Fourier Explorer')
        self.figtitle = self.fig.suptitle('')
        self.ax_img = self.fig.add_subplot(1, 2, 1)
        self.ax_fft = self.fig.add_subplot(1, 2, 2)

        # Show the filtered image
        if cmap is None:
            self.im_img = self.ax_img.imshow(self.img, cmap=matplotlib.cm.gray)
        else:
            self.im_img = self.ax_img.imshow(self.img, cmap=cmap)
        self.ax_img.set_title('Filtered image')
        self.ax_img.axis('off')

        self.update_img()

        # Show the Fourier transform
        logabsfftimg = np.fft.fftshift(np.log(np.abs(self.fimg)))
        self.ax_fft.imshow(logabsfftimg, vmin=1, vmax=logabsfftimg[logabsfftimg!=logabsfftimg[0, 0]].max(), cmap=matplotlib.cm.gray)
        self.ax_fft.set_title('FFT')
        self.ax_fft.axis('off')

        plt.subplots_adjust(left=.05, bottom=.05, right=.95, top=.92, wspace=.05, hspace=.05)

        # Add the alpha mask
        self.alpha = 180.
        alpha_mask = np.zeros(self.shape + (4,), dtype='uint8')
        alpha_mask[:, :, 2] = 255 # blue channel
        self.im_alpha = self.ax_fft.imshow(alpha_mask)
        self.alpha_mask = self.im_alpha._A
        self.update_alpha_mask()

        self.useblit = useblit
        if self.useblit:
            self.update_background()

        # Prepare pen
        self.pen_type = 'gauss'
        self.pen_size = 5
        self._r = np.sqrt(compute_q(self.shape)[2])
        self.update_pen()

        # Connect the event handlers
        self.draw_mode = None
        self.eventpress = None
        self.fig.canvas.mpl_connect('button_press_event', self.on_button_press)
        self.fig.canvas.mpl_connect('button_release_event', self.on_button_release)
        self.fig.canvas.mpl_connect('motion_notify_event', self.on_motion_notify)
        self.fig.canvas.mpl_connect('key_press_event', self.on_key_press)

        self.useblit = useblit
        if self.useblit:
            self.update_background()

        plt.show()

    def symmetrize_mask(self):
        """\
        Ensures that the mask remains centro-symmetric.
        """
        self.mask += np.roll(np.roll(np.flipud(np.fliplr(self.mask)), 1, axis=0), 1, axis=1)
        self.mask = self.mask.clip(0, 1)

    def symmetrize(self, mask):
        """\
        Ensures that the given mask remains centro-symmetric.
        """
        m1 = np.roll(np.roll(np.flipud(np.fliplr(mask)), 1, axis=0), 1, axis=1)
        m1[mask > m1] = mask[mask>m1]
        return m1.clip(0., 1.)

    def update_alpha_mask(self, mask=None):
        """\
        Ensures that the alpha mask is consistent with the mask.
        """
        if mask is None: mask = self.mask
        self.alpha_mask[:, :, 3] = (self.alpha * np.fft.fftshift(mask)).astype(np.uint8)
        self.im_alpha._imcache = None  # This ensures the mask will be redrawn

    def update_mask(self):
        """\
        Ensures that the alpha mask is consistent with the mask.
        """
        self.mask = (np.fft.ifftshift(self.alpha_mask[:, :, 3]) / self.alpha).clip(0, 1).astype(float)

    def update_img(self):
        """\
        Computes the filtered image consistent with the mask.
        """
        self.filt_img = np.real(np.fft.ifftn(self.mask * self.fimg))
        self.im_img.set_data(self.filt_img)
        self.im_img.set_clim((self.filt_img.min(), self.filt_img.max()))

    def update_pen(self):
        """\
        Computes a disc-shape pen.
        """
        if self.pen_type == 'disc':
            self.pen = (self._r <= self.pen_size/2.).astype(float)
        elif self.pen_type == 'gauss':
            self.pen = np.exp(-(self._r**2)/self.pen_size**2)
            self.pen[self.pen < 1e-3] = 0.
        self.figtitle.set_text("Pen type: %s (hit the spacebar to toggle). \nPen size: %d (hit +/- to change). Hit 'r' to flip mask" % (self.pen_type, self.pen_size))
        self.fig.canvas.draw_idle()

    def on_button_press(self, evt):
        """\
        Register a click.
        """
        if evt.inaxes == self.ax_fft and evt.button == 1 and not self.fig.canvas.manager.toolbar.mode:
            # button press in the FFT window
            self.eventpress = evt
            x, y = int(np.round(evt.xdata)), int(np.round(evt.ydata))
            self.draw_mode = (2*self.alpha_mask[y, x, 3] < self.alpha)
            self.new_mask = np.zeros_like(self.mask)
            if self.useblit:
                self.update_background()
            self.on_motion_notify(evt)
        elif evt.inaxes == self.ax_img and evt.button == 1:
            # button press in the image window
            pass
            
    def on_motion_notify(self, evt):
        if self.eventpress is None or evt.inaxes != self.ax_fft or self.draw_mode is None:
            return
        x, y = int(np.round(evt.xdata)), int(np.round(evt.ydata))
        fy = (y - (self.shape[0]+1)//2) % self.shape[0]
        fx = (x - (self.shape[1]+1)//2) % self.shape[1]
        new_mask_incr = self.symmetrize(np.roll(np.roll(self.pen, fy, axis=0), fx, axis=1))
        self.new_mask[new_mask_incr > self.new_mask] = new_mask_incr[new_mask_incr > self.new_mask]
        if self.draw_mode:
            self.update_alpha_mask((self.mask + self.new_mask).clip(0., 1.))
        else:
            self.update_alpha_mask((self.mask - self.new_mask).clip(0., 1.))
        self.im_alpha._imcache = None  # This ensures the mask will be redrawn
        self.update_fft()
        #self.update_img()
        self.update_all()

    def on_button_release(self, evt):
        if self.eventpress is None or evt.button != 1:
            return
        if self.draw_mode:
            self.mask = (self.mask + self.new_mask).clip(0., 1.)
        else:
            self.mask = (self.mask - self.new_mask).clip(0., 1.)
        self.draw_mode = None
        self.update_img()
        self.update_all()
        self.eventpress = None

    def on_key_press(self, evt):
        if evt.key in ['=', '+']:
            self.pen_size += 1
            self.update_pen()
        elif evt.key in ['-', '_']:
            self.pen_size = max(1, self.pen_size-1)
            self.update_pen()
        elif evt.key in [' ']:
            self.pen_type = 'gauss' if self.pen_type == 'disc' else 'disc'
            self.update_pen()
        elif evt.key in ['r', 'R']:
            self.mask = 1-self.mask
            self.update_alpha_mask()
            self.update_img()
            self.update_fft()
            self.update_all()

    def update_background(self):
        """\
        Force an update of the background
        """
        if self.useblit:
            self.im_alpha.set_visible(False)
            self.fig.canvas.draw()
            self.background = self.fig.canvas.copy_from_bbox(self.ax_fft.bbox)
            self.im_alpha.set_visible(True)
            self.fig.canvas.draw()

    def update_fft(self):
        if self.useblit:
            if self.background is not None:
                self.fig.canvas.restore_region(self.background)
            self.ax_fft.draw_artist(self.im_alpha)
            self.fig.canvas.blit(self.ax_fft.bbox)
        else:
            self.fig.canvas.draw_idle()

    def update_all(self):
        if self.useblit:
            self.update_background()
        self.fig.canvas.draw_idle()


if __name__ == "__main__":
    import sys
    from imageio import imread
    
    if len(sys.argv) == 1:
        # load test image
        image_file = 'logo2.png'
    else:
        image_file = sys.argv[1]
    img = imread(image_file)/255.
    if img.ndim == 3:
        img = img.mean(axis=-1)
    print(img.shape)

    # start application
    F = FExplorer(img)
